"""
'sven.exc': svn-related exceptions
"""

class NotAFile(IOError):
    def __init__(self, uri):
        IOError.__init__(self, 21, "Is a directory", uri)

class NotADirectory(IOError):
    def __init__(self, uri):
        IOError.__init__(self, 20, "Not a directory", uri)

class NoSuchResource(IOError):
    def __init__(self, uri):
        IOError.__init__(self, 2, "No such file or directory", uri)

class ResourceUnchanged(Exception):
    def __init__(self, uri, last_change):
        Exception.__init__(self, "Resource '%s' unchanged since %d" % (uri, last_change))
        self.last_change = last_change

class ResourceChanged(Exception):
    def __init__(self, uri):
        Exception.__init__(self, "Resource '%s' is out of date" % uri)
        self.uri = uri

class FutureRevision(Exception):
    def __init__(self, rev):
        Exception.__init__(self, "No such revision %s yet in the repository" % rev)
        self.rev = rev



from operator import attrgetter
import os
import pysvn
from sven.exc import *

class BaseSvnAccess(object):
    def __init__(self, checkout_dir,
                 config_location=None,
                 default_commit_message=None,
                 path_fixer=None):

        if config_location and not config_location.startswith('/'):
            config_location = os.path.join(checkout_dir, config_location)

        if config_location and config_location.endswith('/'):
            config_location = config_location.rstrip('/')

        self.config_location = config_location or ''

        #os.chdir(checkout_dir)
        self.checkout_dir = checkout_dir

        self.default_message = default_commit_message or "foom"

        self.path_fixer = path_fixer

    def normalized(self, path):
        """
        normalizes a path

        >>> normalized('/my/path/') == normalized('my/path') == 'my/path'
        True

        optionally a `path_fixer` may be applied, if set on the class. 
        the path will still be guaranteed to have no leading or trailing slashes.
        """
        path = path.strip('/')
        if self.path_fixer:
            path = self.path_fixer(path)
        return path.strip('/')

    @property
    def client(self):
        client = pysvn.Client(self.config_location)
        if self.config_location.startswith(self.checkout_dir):
            client.add(self.config_location[len(self.checkout_dir):])
        client.exception_style = 1
        #client.callback_notify = notify
        return client

    def last_changed_rev(self, uri, rev=None):
        uri = self.normalized(uri)
        
        absolute_uri = '/'.join((self.checkout_dir, uri))

        try:
            log = self.client.log(absolute_uri,
                                  discover_changed_paths=True)
        except pysvn.ClientError, e:
            if e[1][0][1] == 155007: # not a working copy
                raise NoSuchResource(uri)
            if e[1][0][1] == 160013: # file not found
                raise NoSuchResource(uri)
            if e[1][0][1] == 150000: # not under version control
                raise NoSuchResource(uri)
            raise
            
        if rev is not None:
            rev = pysvn.Revision(pysvn.opt_revision_kind.number, rev)
            try:
                info = self.client.info2(absolute_uri, rev)
            except pysvn.ClientError, e:
                if e[1][0][1] == 150004: # has no URL
                    raise NoSuchResource(uri)
                if e[1][0][1] == 155007: # not a working copy
                    raise NoSuchResource(uri)
                raise                
        else:
            try:
                info = self.client.info2(absolute_uri)
            except pysvn.ClientError, e:
                if e[1][0][1] == 200005: # not under version control
                    raise NoSuchResource(uri)
                if e[1][0][1] == 150000: # not under version control
                    raise NoSuchResource(uri)
                if e[1][0][1] == 155007: # not a working copy
                    raise NoSuchResource(uri)
                raise

        last_change = info[0][1].last_changed_rev.number
        

        ##### in case the file's parent directory was moved more recently than the file was changed
        # this is particularly tricky if the sven client is "mounted" at a child directory of the moved path
        # e.g. 
        # touch /foo/bar/baz.txt
        # x = SvnAccess('/foo/bar/')
        # svn mv /foo /fleem
        # x.info('baz.txt')

        log = self.client.log(absolute_uri, discover_changed_paths=True)
        problems = [x for x in log[0]['changed_paths']
                    if x['copyfrom_path']]

        repo_url = info[0][1]['repos_root_URL']
        doc_url = info[0][1]['URL']

        move_revision = None
        for problem in problems:
            base_moved_path = '/'.join((repo_url.rstrip('/'), self.normalized(problem['path']) ))
            assert doc_url.startswith(base_moved_path), "I don't know what this means!\nPlease let me know the circumstances of this error if you hit it: ejucovy+sven@gmail.com"
            move_revision = problem.copyfrom_revision.number

        if move_revision:
            if last_change > move_revision:
                return last_change
            return move_revision +1 # +1 is the oldest revision at which the file exists at its current location.

        return last_change

    def read(self, uri, rev=None):
        """
        Return the raw @type:string data stored in @param:uri
        at @param:rev in some sort of ad-hoc JSON format.
        
        @raise:NotAFile
        @raise:NoSuchResource
        @raise:ResourceUnchanged
        XXX TODO @raise additional information for edge cases of accessing
               a file at a revision when it did not yet exist, or a revision
               when it moved or ceased to exist.
        """
        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        index_template = """<html><head><title></title></head>
<body><div id='content'><ul>%s</ul></div></body></html>"""
        if rev is not None:
            last_change = self.last_changed_rev(uri, rev)
            if last_change < int(rev):
                raise ResourceUnchanged(uri, last_change)

            rev = pysvn.Revision(pysvn.opt_revision_kind.number, rev)
            try:
                return {'body': self.client.cat(absolute_uri, rev),
                        'mimetype': self.mimetype(uri),
                        }
            except pysvn.ClientError, e:
                if e[1][0][1] == 195007: # URL refers to a directory
                    raise NotAFile(uri)
                raise

        try:
            return {'body': file(absolute_uri).read(),
                    'mimetype': self.mimetype(uri),
                    }
        except IOError, e:
            if e.errno == 21:
                raise NotAFile(uri)
            elif e.errno == 2:
                raise NoSuchResource(uri)
            raise

    def ls(self, uri, rev=None):
        """
        Return the listing of contents stored under @param:uri
        as a JSON-listing.

        @raise:NotADirectory
        @raise:ResourceUnchanged
        etc.
        """

        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        # if uri evals to false we're at the repo root
        if uri and not os.path.isdir(absolute_uri):
            raise NotADirectory(uri)

        if rev is not None:
            last_change = self.last_changed_rev(uri, rev)
            if last_change < int(rev):
                raise ResourceUnchanged(uri, last_change)

            rev = pysvn.Revision(pysvn.opt_revision_kind.number, rev)
            contents = self.client.ls(absolute_uri, rev)
        else:
            contents = self.client.ls(absolute_uri)

        contents.sort(key=attrgetter('name'))
        globs = []
        for obj in contents:
            obj_name = obj.name
            if obj_name.startswith(self.checkout_dir):
                obj_name = obj_name[len(self.checkout_dir):]
            glob = dict(href='%s' % obj_name)
            fields = {'id': obj_name}
            if rev is not None:
                fields['version'] = rev.number
            glob['fields'] = fields
            globs.append(glob)
        return globs

    def log(self, uri, rev=None):
        """
        Return the changelog of data stored at or under @param:uri
        as of time @param:rev in JSON-listing format.
        
        @raise:ResourceUnchanged
        etc.
        """

        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        if rev is not None:
            last_change = self.last_changed_rev(uri, rev)
            if last_change < int(rev):
                raise ResourceUnchanged(uri, last_change)

            rev = pysvn.Revision(pysvn.opt_revision_kind.number, rev)
            try:
                log = self.client.log(absolute_uri, rev,
                                      discover_changed_paths=True)
            except pysvn.ClientError, e:
                if e[1][0][1] == 150004: # has no URL
                    raise NoSuchResource(uri)
                raise
        else:
            try:
                log = self.client.log(absolute_uri,
                                      discover_changed_paths=True)
            except pysvn.ClientError, e:
                if e[1][0][1] == 200005: # not under version control
                    raise NoSuchResource(uri)
                raise

        globs = []
        for obj in log:
            href = obj.changed_paths[0].path
            glob = dict(href=href)
            fields = {'id': href,
                      'message': obj.message,
                      'author': obj.author,
                      'version': obj.revision.number}
            glob['fields'] = fields
            globs.append(glob)
        return globs

    def mimetype(self, uri, rev=None):
        uri = self.normalized(uri)
        absolute_uri = '/'.join((
                self.checkout_dir.rstrip('/'),
                uri))

        if rev is not None:
            rev = pysvn.Revision(pysvn.opt_revision_kind.number, rev)
            def propget():
                return self.client.propget('svn:mime-type', absolute_uri, rev)
        else:
            def propget():
                return self.client.propget('svn:mime-type', absolute_uri)
        try:
            properties = propget()
        except pysvn.ClientError, e:
            if e[1][0][1] == 150000: # not under version control
                return ""
            if e[1][0][1] == 200005: # not under version control
                return ""
            if e[1][0][1] == 155007: # not a working copy
                return ""
            raise
    
        return properties.get(absolute_uri)

    def set_mimetype(self, uri, mimetype, msg=None):

        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        self.client.propset('svn:mime-type', mimetype, absolute_uri)
        
        if not msg:
            msg = "Set svn:mime-type property to '%s'" % mimetype
        commit_rev = self.client.checkin([absolute_uri], msg)

        return commit_rev
        
    def write(self, uri, contents, msg=None, mimetype=None, 
              use_newline=True, binary=False):
        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        if os.path.isdir(absolute_uri): # we can't write to a directory
            raise NotAFile(uri)

        parent_dir = '/'.join(uri.split('/')[:-1])
        absolute_parent_dir = '/'.join((self.checkout_dir, parent_dir))

        if parent_dir and not os.path.exists(absolute_parent_dir):
            path = '/'.join(uri.split('/')[:-1])
            os.makedirs('/'.join((self.checkout_dir, path)))

            path = path.split('/')
            success = False
            for i in range(len(path)):
                root_to_add = '/'.join(path[:i+1])
                try: 
                    self.client.add('/'.join((self.checkout_dir, root_to_add)))
                    success = True
                except pysvn.ClientError, e: #XXX TODO: typecheck this error
                    if not success:
                        # the resource is already under version control
                        # but not because we just added it. keep walking
                        # up the chain to find something that isn't yet
                        # under version control.
                        continue
                    
                    # the resource is already under version control
                    # because we just added its parent directory.
                    # walk one step down the chain to get back to the
                    # root directory that we just added.
                    root_to_add = '/'.join(path[:i])
                    break
            self.client.checkin(['/'.join((self.checkout_dir, root_to_add))],
                                "auto-creating directories")
            path_to_update = '/'.join([self.checkout_dir] + path)
            self.client.update(path_to_update)

        mode = 'w'
        if binary is True:
            use_newline = False
            mode = 'wb'
        f = file(absolute_uri, mode)
        if use_newline and not contents.endswith('\n'):
            contents += '\n'
        f.write(contents)
        f.close()

        try:
            self.client.add(absolute_uri)
        except pysvn.ClientError, e:
            if e[1][0][1] == 150002: # already under version control
                pass
            else: # i don't know what else this would be! better not make any decisions!
                raise

        if not msg: # wish we could just do `if msg is None`, but we can't.
            msg = self.default_message

        if mimetype:
            self.client.propset('svn:mime-type', mimetype, absolute_uri)

        try:
            commit_rev = self.client.checkin([absolute_uri], msg)
            return commit_rev
        except pysvn.ClientError, e:
            if e[1][0][1] == 160028: # file is out of date
                # we roll back our attempted changes and let the caller deal with merges
                self.client.revert(absolute_uri)
                raise ResourceChanged(uri)
            else: # i don't know what else this would be! better not make any decisions!
                raise



class SvnAccessWriteUpdateHandler(BaseSvnAccess):
    def set_mimetype(self, uri, mimetype, msg=None,
                     update_before_write=True,
                     update_after_write=True):
        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        if update_before_write:
            self.client.update(self.checkout_dir)

        BaseSvnAccess.set_mimetype(self, uri, mimetype, msg)

        if update_after_write:
            self.client.update(self.checkout_dir)

    def write(self, uri, contents, msg=None, mimetype=None,
              update_before_write=True,
              update_after_write=True):
        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        if update_before_write:
            self.client.update(self.checkout_dir)

        result = BaseSvnAccess.write(self, uri, contents, msg, mimetype)

        if update_after_write:
            self.client.update(self.checkout_dir)

        return result

#bbb
#deprecate in 0.5
# XXX????
SvnAccess = SvnAccessWriteUpdateHandler

class SvnAccessEventEmitter(SvnAccess):
    def __init__(self, *args, **kw):
        SvnAccess.__init__(self, *args, **kw)
        self.listeners = []

    def add_listener(self, callback):
        self.listeners.append(callback)

    def set_mimetype(self, uri, mimetype, msg=None):

        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        pre_rev = self.client.info2(absolute_uri)[0][1].rev
        post_rev = SvnAccess.set_mimetype(self, uri, mimetype, msg)

        for callback in self.listeners:
            callback(absolute_uri, contents, msg, mimetype, (pre_rev, post_rev))

    def write(self, uri, contents, msg=None, mimetype=None):

        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        pre_rev = self.client.info2(absolute_uri)[0][1].rev
        post_rev = SvnAccess.write(self, uri, contents, msg, mimetype)
        for callback in self.listeners:
            callback(uri, contents, msg, mimetype, (pre_rev, post_rev))

if __name__ == '__main__':
    import doctest
    doctest.testfile('doctest.txt')



from operator import attrgetter
import os
import pysvn
from sven.exc import *




class FSAccess(object):
    def __init__(self, checkout_dir):

        #if config_location and not config_location.startswith('/'):
        #    config_location = os.path.join(checkout_dir, config_location)

        #if config_location and config_location.endswith('/'):
        #    config_location = config_location.rstrip('/')

        #self.config_location = config_location or ''

        #os.chdir(checkout_dir)
        self.checkout_dir = checkout_dir

        #self.default_message = default_commit_message or "foom"


    def read(self, uri):
        """
        Return the raw @type:string data stored in @param:uri
        at @param:rev in some sort of ad-hoc JSON format.
        
        @raise:NotAFile
        @raise:NoSuchResource
        XXX TODO @raise additional information for edge cases of accessing
               a file at a revision when it did not yet exist, or a revision
               when it moved or ceased to exist.
        """
        uri = uri.strip('/')
        absolute_uri = '/'.join((self.checkout_dir, uri))

        try:
            return {'body': file(absolute_uri).read(),
                    'kind': None,
                    }
        except IOError, e:
            if e.errno == 21:
                raise NotAFile(uri)
            elif e.errno == 2:
                raise NoSuchResource(uri)
            raise

    def write(self, uri, contents):
        uri = uri.strip('/')
        absolute_uri = '/'.join((self.checkout_dir, uri))

        if os.path.isdir(absolute_uri): # we can't write to a directory
            raise NotAFile(uri)

        parent_dir = '/'.join(uri.split('/')[:-1])
        absolute_parent_dir = '/'.join((self.checkout_dir, parent_dir))

        if parent_dir and not os.path.exists(absolute_parent_dir):
            path = '/'.join(uri.split('/')[:-1])
            os.makedirs('/'.join((self.checkout_dir, path)))

            path = path.split('/')
            success = False
            for i in range(len(path)):
                root_to_add = '/'.join(path[:i+1])

        f = file(absolute_uri, 'w')
        print >> f, contents
        f.close()

if __name__ == '__main__':
    import doctest
    doctest.testfile('doctest.txt')



from operator import attrgetter
import os
from bzrlib import workingtree
from bzrlib.errors import NoSuchRevision, BoundBranchOutOfDate, ConflictsInTree
from bzrlib.inventory import InventoryDirectory
from sven.exc import *

# >>> x=workingtree.WorkingTree.open('bar')
# >>> x.branch.lock_read()
# >>> [x.branch.revision_id_to_revno(a) for a in x.[x.path2id('bar')]]
# [1, 2]

# x = BzrAccess('bar').client

# much help from http://code.google.com/p/django-rcsfield/source/browse/trunk/rcsfield/backends/bzr.py

class BzrAccess(object):
    def __init__(self, checkout_dir,
                 config_location=None,
                 default_commit_message=None,
                 path_fixer=None):

        if config_location and not config_location.startswith('/'):
            config_location = os.path.join(checkout_dir, config_location)

        if config_location and config_location.endswith('/'):
            config_location = config_location.rstrip('/')

        self.config_location = config_location or ''

        #os.chdir(checkout_dir)
        self.checkout_dir = checkout_dir

        self.default_message = default_commit_message or "foom"

        self.path_fixer = path_fixer

    def normalized(self, path):
        """
        normalizes a path

        >>> normalized('/my/path/') == normalized('my/path') == 'my/path'
        True

        optionally a `path_fixer` may be applied, if set on the class. 
        the path will still be guaranteed to have no leading or trailing slashes.
        """
        path = path.strip('/')
        if self.path_fixer:
            path = self.path_fixer(path)
        return path.strip('/')

    @property
    def client(self):
        client = workingtree.WorkingTree.open(self.checkout_dir)
        return client

    def iter_content(self, uri):
        """
        Yield the (rev, content) at uri in each revision it changed,
        from most recent to first
        """
        revisions = self.revisions(uri)
        for rev in revisions:
            yield rev, self.read(uri, rev=rev)

    def revisions(self, uri):
        """
        revisions at which this file changed
        """
        uri = self.normalized(uri)

        x = self.client
        path = x.path2id(uri)
        if not path:
            raise NoSuchResource(uri)

        from bzrlib.log import LogFormatter

        def get_formatter(lst):
            class ListLogFormatter(LogFormatter):
                def __init__(self, *args, **kw):
                    LogFormatter.__init__(self, *args, **kw)
                    self._loglist = lst
                def log_revision(self, revision):
                    self._loglist.append(int(revision.revno))
            return ListLogFormatter

        from bzrlib.builtins import cmd_log
        from bzrlib.revisionspec import RevisionSpec

        log = cmd_log()
        log.outf = None

        foo = []

        rev = None

        absolute_uri = os.path.join(self.checkout_dir, uri)

        log.run(file_list=[absolute_uri],
                revision=rev and [RevisionSpec.from_string(str(rev))] or None,
                log_format = get_formatter(foo),
                )

        return foo

    def last_changed_rev(self, uri, rev=None):
        """
        @raise:NoSuchResource
        """
        uri = self.normalized(uri)

        changes = self.revisions(uri)

        if changes is None:
            return None

        if rev is None:
            return changes[0]

        rev = int(rev)

        for revno in changes:
            if rev < revno:
                continue
            return revno

        raise NoSuchResource(uri)

    def read(self, uri, rev=None):
        """
        Return the raw @type:string data stored in @param:uri
        at @param:rev in some sort of ad-hoc JSON format.
        
        @raise:NotAFile
        @raise:NoSuchResource
        @raise:ResourceUnchanged
        XXX TODO @raise additional information for edge cases of accessing
               a file at a revision when it did not yet exist, or a revision
               when it moved or ceased to exist.
        """
        uri = self.normalized(uri)
                
        x = self.client

        if rev is not None:
            rev = int(rev)

        if rev is not None:            
            try:
                rev_id = x.branch.get_rev_id(int(rev))
            except NoSuchRevision, e:
                raise FutureRevision(int(rev))
            x = x.branch.repository.revision_tree(rev_id)

        path = x.path2id(uri)
        if not path:
            raise NoSuchResource(uri)

        if rev is not None:
            last_change = self.last_changed_rev(uri, rev=rev)
            if last_change < rev:
                raise ResourceUnchanged(uri, last_change)

        x.lock_read()

        files = dict([(i[0], i[2]) for i in list(x.list_files())])
        
        if uri in files and files[uri] == 'directory':
            raise NotAFile(uri)

        try:
            data = x.get_file(path)
        except IOError, e:
            x.unlock()
            if e.errno == 21:
                raise NotAFile(uri)
            raise

        x.unlock()

        data = data.read()
        return data

    def ls(self, uri, rev=None):
        """
        Return the listing of contents stored under @param:uri
        as a JSON-listing.

        @raise:NotADirectory
        @raise:ResourceUnchanged
        etc.
        """
                
        x = self.client

        uri = self.normalized(uri)

        x.lock_read()

        if rev is not None:
            rev = int(rev)

            rev_id = x.branch.get_rev_id(int(rev))
            y = x.branch.repository.revision_tree(rev_id)
            x.unlock()
            x = y
            x.lock_read()

        inv = x.inventory

        path = x.path2id(uri)
        if path is None:
            raise NoSuchResource(uri)

        dir = inv[path]

        x.unlock()

        if not isinstance(dir, InventoryDirectory):
            raise NotADirectory(uri)

        if rev is not None:
            last_change = self.last_changed_rev(uri, rev=rev)
            if last_change < rev:
                raise ResourceUnchanged(uri, last_change)

        contents = ["%s/%s" % (uri, key) for key in dir.children.keys()]

        globs = []
        for obj in contents:
            glob = dict(href=obj)
            fields = {'id': obj}
            if rev is not None:
                fields['version'] = rev
            glob['fields'] = fields
            globs.append(glob)
        return globs

    def proplist(self, uri):
        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        if os.path.isdir(absolute_uri):
            path = '/'.join((uri, '.sven-meta'))
            try:
                return [i['fields']['id']
                        [len(uri)+len('.sven-meta/.')+1:] 
                        for i in self.ls(path)]
            except NoSuchResource:
                return []
        else:
            path = '/.sven-meta'
            return [i['fields']['id'][len('.sven-meta/.'):] 
                    for i in self.ls(path)]
            
    def log(self, uri, rev=None):
        """
        Return the changelog of data stored at or under @param:uri
        as of time @param:rev in JSON-listing format.
        
        @raise:ResourceUnchanged
        etc.
        """

        from bzrlib.log import LogFormatter

        def get_formatter(lst):
            class ListLogFormatter(LogFormatter):
                def __init__(self, *args, **kw):
                    LogFormatter.__init__(self, *args, **kw)
                    self._loglist = lst
                def log_revision(self, revision):
                    revno = int(revision.revno)
                    author = revision.rev.committer  # @@@ what about get_apparent_authors?
                    message = revision.rev.message.rstrip('\r\n')
                    timestamp = revision.rev.timestamp
                    self._loglist.append(dict(version=revno,
                                              author=author,
                                              message=message,
                                              timestamp=timestamp,
                                              id=uri,
                                              revprops=revision.rev.properties))
            return ListLogFormatter

        from bzrlib.builtins import cmd_log
        from bzrlib.revisionspec import RevisionSpec

        log = cmd_log()
        log.outf = None

        foo = []

        absolute_uri = os.path.join(self.checkout_dir, self.normalized(uri))

        log.run(file_list=[absolute_uri],
                revision=rev and [RevisionSpec.from_string("1"),
                                  RevisionSpec.from_string(str(rev))] or None,
                log_format = get_formatter(foo),
                )

        return [dict(href=uri, fields=i) for i in foo]

    def mimetype(self, uri, rev=None):
        return self.propget(uri, 'mimetype', rev=rev)

    def propget(self, uri, prop, rev=None):
        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        if os.path.isdir(absolute_uri):
            return self._dir_prop(uri, prop, rev=rev)
        return self._file_prop(uri, prop, rev=rev)        

    def _dir_prop(self, uri, prop, rev=None):
        uri = self.normalized(uri)

        if not uri:
            return None

        absolute_props_uri = '/'.join((self.checkout_dir, uri, '.sven-meta'))

        if not os.path.isdir(absolute_props_uri):
            return None
        res = self.read('/'.join((uri, '.sven-meta', '.%s' % prop)), rev=rev)
        if res: return res['body']
        return res

    def _file_prop(self, uri, prop, rev=None):
        uri = self.normalized(uri)
        props_uri = '/'.join((
                '.sven-meta/.%s' % prop, uri))
        
        if not uri:
            raise RuntimeError("Can't do that")

        try:
            res = self.read(props_uri, rev=rev)
        except NoSuchResource, e:
            return None
        if res: return res['body']
        return res

    def _dir_propset(self, uri, prop, val, msg=None):
        uri = self.normalized(uri)

        return self.write('/'.join((uri, '.sven-meta/.%s' % prop)),
                          val, use_newline=False, commit=False)

    def _file_propset(self, uri, prop, val, msg=None):
        uri = self.normalized(uri)

        return self.write('/'.join(('.sven-meta/.%s' % prop, uri)),
                          val, use_newline=False, commit=False)

    def propset(self, uri, prop, val, msg=None):
        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        if os.path.isdir(absolute_uri):
            return self._dir_propset(uri, prop, val, msg=msg)
        return self._file_propset(uri, prop, val, msg=msg)
        
    def set_mimetype(self, uri, mimetype, msg=None):
        return self.propset(uri, 'mimetype', mimetype, msg=msg)

    def commit(self, uri, msg=None):
        x = self.client
        if msg is None:
            msg = self.default_message
        try:
            rev_id = x.commit(message=msg)
        except (BoundBranchOutOfDate, ConflictsInTree), e:
            raise ResourceChanged(uri)

        return R(x.branch.revision_id_to_revno(rev_id))

    def write(self, uri, contents, msg=None, mimetype=None,
              use_newline=True, binary=False,
              commit=True,
              metadata=None, revprops=None):
        uri = self.normalized(uri)
        absolute_uri = '/'.join((self.checkout_dir, uri))

        if os.path.isdir(absolute_uri): # we can't write to a directory
            raise NotAFile(uri)

        parent_dir = os.path.dirname(uri)
        absolute_parent_dir = '/'.join((self.checkout_dir, parent_dir))

        x = self.client

        if parent_dir and not os.path.exists(absolute_parent_dir):
            os.makedirs(absolute_parent_dir)
            x.smart_add([absolute_parent_dir])
            #x.commit("Auto-creating directories")

        mode = 'w'
        if binary is True:
            mode = 'wb'
            use_newline = False
        f = file(absolute_uri, mode)

        if use_newline and not contents.endswith('\n'):
            contents += '\n'
        f.write(contents)
        f.close()

        x.add([uri])

        if not msg: # wish we could just do `if msg is None`, but we can't.
            msg = self.default_message

        metadata = metadata or {}
        if mimetype is not None:
            assert 'mimetype' not in metadata
            metadata['mimetype'] = mimetype

        for key, val in metadata.items():
            self.propset(uri, key, val, msg=msg)

        if not commit:
            return

        try:
            rev_id = x.commit(message=msg, revprops=revprops)
        except (BoundBranchOutOfDate, ConflictsInTree), e:
            raise ResourceChanged(uri)

        return R(x.branch.revision_id_to_revno(rev_id))

class UpdatingBzrAccess(BzrAccess):
    def __init__(self, checkout_dir, config_location=None,
                 default_commit_message=None, path_fixer=None,
                 update_before_write=True, update_after_write=True):
        BzrAccess.__init__(self, checkout_dir, config_location,
                           default_commit_message, path_fixer)
        self.update_before_write = update_before_write
        self.update_after_write = update_after_write

    def write(self, *args, **kw):
        override = kw.get('update_before_write', None)
        if override is not None:
            before_write = override
            del kw['update_before_write']
        else:
            before_write = self.update_before_write
        if before_write:
            self.client.update()
        res = BzrAccess.write(self, *args, **kw)
        if self.update_after_write:
            self.client.update()
        return res

class Revision(object):
    def __init__(self, n):
        self.n = n
    def __repr__(self):
        return "<Revision kind=number %d>" % self.n
R = Revision

if __name__ == '__main__':
    import doctest
    doctest.testfile('bzr.txt')



"""
implements a `path_fixer` for the backend
"""

from datetime import date
class DateLayoutPathFixer(object):
    """
    a bloggish directory structure

    assuming today is 2009-12-22::
    >>> fixed_path = DateLayoutPathFixer()
    >>> fixed_path("/my/post.html")
    '2009/12/22/my/post.html'

    you can set the date and a timedelta::
    >>> fixed_path = DateLayoutPathFixer(from_date=datetime.datetime(2009, 12, 20), delta=datetime.timedelta(4))
    >>> fixed_path("/my/post.html")
    '2009/12/24/my/post.html'
    """
    def __init__(self, from_date=None, delta=None):
        self.from_date = from_date
        self.delta = delta
    def __call__(self, path):
        from_date = self.from_date or date.today()
        if self.delta:
            from_date += delta
        prefix = from_date.strftime("%Y/%m/%d")
        return "%s/%s" % (prefix, path)




from setuptools import setup, find_packages
import sys, os

version = '0.8.1'

long_description = open('README.txt').read()
new_in_this_version = open('changes/changes.txt').read()
history = open('changes/history.txt').read()

long_description = """
%s

New in this version
-------------------

%s

History
-------

%s
""" % (long_description, new_in_this_version, history)

setup(name='sven',
      version=version,
      description="sven is a document-oriented programming library that helps you put content in a version-controlled document repository",
      long_description=long_description,
      classifiers=[],
      keywords='',
      author='Ethan Jucovy & Jeff Hammel',
      author_email='ejucovy@gmail.com',
      url='',
      license='GPL',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
      ],
      entry_points="""
      """,
      )
